/* TCPGet.c : a file-receive mechanism that only requires a TCP/IP stack to work
   Compiles for Win32 with MSVC++ 6.0, using the WinSock2 API
   Compiler for AmigaOS with SAS/C 6.50, using the Miami SDK

   by Mikael Kalms (mikael@kalms.org), 2003-03-03

   Be careful when using this program: When the client is active, anyone can write
    to its harddrive. Also, it will overwrite existing files without any questions!
*/

#ifdef _WIN32

#include <winsock2.h>
#include <windows.h>

#define GETSOCKETERROR WSAGetLastError

#else

#include <proto/socket.h>
#include <sys/socket.h>

#define GETSOCKETERROR Errno

typedef long SOCKET;

#define SOCKET_ERROR -1
#define INVALID_SOCKET -1

#endif

#include <stdio.h>

#define BLKSIZE 1024

SOCKET listenerMainSocket;
SOCKET listenerCommunicationsSocket;

int verbose = 1;

char fileName[256];

int createListener(void)
{
	int ret;
	struct sockaddr_in listenerAddr;
	int listenerHostPort = 17501;

	memset(&listenerAddr, 0, sizeof(listenerAddr));

#ifndef _WIN32
	listenerAddr.sin_len			= 0;
#endif
	listenerAddr.sin_family			= AF_INET;
	listenerAddr.sin_port			= htons(listenerHostPort);
	listenerAddr.sin_addr.s_addr	= htonl(INADDR_ANY);

	if ((listenerMainSocket = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET)
	{
		fprintf(stderr, "socket: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("socket OK\n");

	if ((ret = bind(listenerMainSocket, (struct sockaddr*)&listenerAddr, sizeof(struct sockaddr))) == SOCKET_ERROR)
	{
		fprintf(stderr, "bind: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("bind OK\n");

	if ((ret = listen(listenerMainSocket, SOMAXCONN)) == SOCKET_ERROR)
	{
		fprintf(stderr, "connect: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("listener OK\n");

	return 1;
}

void destroyListener(void)
{
#ifdef _WIN32
	shutdown(listenerMainSocket, SD_BOTH);

	closesocket(listenerMainSocket);
#else
	CloseSocket(listenerMainSocket);
#endif
}

void acceptFileSocket(void)
{
	listenerCommunicationsSocket = accept(listenerMainSocket, 0, 0);
}

void closeFileSocket(void)
{
#ifdef _WIN32
	shutdown(listenerCommunicationsSocket, SD_BOTH);

	closesocket(listenerCommunicationsSocket);
#else
	CloseSocket(listenerCommunicationsSocket);
#endif
}


int receiveBlock(void* buf, int size)
{
	int ret;

	if ((ret = recv(listenerCommunicationsSocket, buf, size, 0)) == SOCKET_ERROR)
	{
		fprintf(stderr, "recv: %d\n", GETSOCKETERROR());
		return -1;
	}

	if (verbose >= 2)
		printf(".");

	return ret;
}

int main(int argc, char** argv)
{
#ifdef _WIN32
	WSADATA data;
#endif

	int ret;
	int arg;
	char buf[BLKSIZE];
	int help = 0;

	for (arg = 1; arg < argc; arg++)
	{
		if (!stricmp(argv[arg], "-q") || !stricmp(argv[arg], "--quiet"))
			verbose = 0;
		else if (!stricmp(argv[arg], "-v") || !stricmp(argv[arg], "--verbose"))
			verbose = 2;
		else if (!stricmp(argv[arg], "-h") || !stricmp(argv[arg], "--help"))
			help = 1;
	}

	if (help)
	{
		printf("TCPGet: File receive client using plain TCP/IP\n");
		printf("By Mikael Kalms (mikael@kalms.org) 2003-03-03\n");
		printf("Usage: tcpput [options] <source filename> <destination host>:[destination filename]\n");
		printf("Options:\n");
		printf("         -h,--help       Display this help text\n");
		printf("         -q,--quiet      Only output error messages\n");
		printf("         -v,--verbose    Output debug info\n");
		return 0;
	}

#ifdef _WIN32
	if ((ret = WSAStartup( MAKEWORD( 2, 2 ), &data )) == SOCKET_ERROR)
	{
		fprintf(stderr, "WSAStartup: %d\n", WSAGetLastError());
		return 1;
	}

	if (verbose >= 2)
		printf("WSAStartup OK\n");
#endif

	if (!createListener())
		return 1;

	if (verbose >= 2)
		printf("createListener() OK\n");

	if (verbose >= 1)
		printf("TCPGet client is ready... (quit with Ctrl-C)\n");

	while (1)
	{
		unsigned long fileNameLen;
		unsigned long fileSize;
		FILE* f;
		unsigned int pos;

		if (verbose >= 2)
			printf("Waiting for new connection...\n");

		acceptFileSocket();

		if (verbose >= 2)
			printf("processListenerAccept() OK\n");

		memset(fileName, 0, sizeof(fileName));

		if (!receiveBlock(&fileNameLen, sizeof(fileNameLen))
			|| !receiveBlock(&fileSize, sizeof(fileSize))
			|| !receiveBlock(&fileName, ntohl(fileNameLen)))
		{
			if (verbose >= 1)
				printf("Connection closed by foreign host\n");
			closeFileSocket();
			continue;
		}

		fileSize = ntohl(fileSize);

		if (verbose >= 1)
			printf("Receiving %d bytes as %s...", fileSize, fileName);

		if (!(f = fopen(fileName, "wb")))
		{
			fprintf(stderr, "Unable to open %s\n", fileName);
			closeFileSocket();
			continue;
		}

		for (pos = 0; pos < fileSize; )
		{
			unsigned int blkSize = BLKSIZE;
			unsigned receivedSize;
			if (pos + blkSize > fileSize)
				blkSize = fileSize - pos;

			receivedSize = receiveBlock(buf, blkSize);

			if (receivedSize < 0)
			{
				break;
			}

			if (receivedSize)
			{
				fwrite(buf, receivedSize, 1, f);
				pos += receivedSize;
			}
		}

		if (verbose >= 1)
			printf("done.\n");

		fclose(f);
		closeFileSocket();
	}

	return 0;
}
