/* TCPPut.c : a file-send mechanism that only requires a TCP/IP stack to work
   Compiles for Win32 with MSVC++ 6.0, using the WinSock2 API
   Compiler for AmigaOS with SAS/C 6.50, using the Miami SDK

   by Mikael Kalms (mikael@kalms.org), 2003-03-03
*/

#ifdef _WIN32

#include <winsock2.h>
#include <windows.h>

#define GETSOCKETERROR WSAGetLastError

#else

#include <string.h>

#include <netdb.h>
#include <proto/socket.h>
#include <sys/socket.h>

#define GETSOCKETERROR Errno

typedef long SOCKET;

#define SOCKET_ERROR -1
#define INVALID_SOCKET -1

#endif


#include <stdio.h>

#define BLKSIZE 1024


char inputName[256];
char outputHostName[256];
char outputName[256];


SOCKET speakerCommunicationsSocket;

int verbose = 1;

int createSpeaker(void)
{
	struct sockaddr_in addr;
	int ret;
	int outputHostPort = 17501;
	unsigned long ip;
	struct hostent* hostEnts;

	if (verbose >= 2)
		printf("Attempting to connect to %s:%d...\n", outputHostName, outputHostPort);

	if (!(hostEnts = gethostbyname(outputHostName)))
	{
		if ((ip = inet_addr(outputHostName)) == INADDR_NONE)
		{
			fprintf(stderr, "%s: No such host\n", outputHostName);
			return 0;
		}
	}
	else
		ip = *((unsigned long*) hostEnts->h_addr);

	memset(&addr, 0, sizeof(addr));

#ifndef _WIN32
	addr.sin_len = 0;
#endif
	addr.sin_family			= AF_INET;
	addr.sin_port			= htons(outputHostPort);
	addr.sin_addr.s_addr	= ip;

	if ((speakerCommunicationsSocket = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET)
	{
		fprintf(stderr, "socket: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("socket OK\n");

	if ((ret = connect(speakerCommunicationsSocket, (struct sockaddr*)&addr, sizeof(struct sockaddr))) == SOCKET_ERROR)
	{
		fprintf(stderr, "connect: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("connect OK\n");

	return 1;
}

void destroySpeaker(void)
{
#ifdef _WIN32
	shutdown(speakerCommunicationsSocket, SD_SEND);

	closesocket(speakerCommunicationsSocket);
#else
	CloseSocket(speakerCommunicationsSocket);
#endif
}

int sendBlk(void* buf, int blkSize)
{
	int ret;
	if ((ret = send(speakerCommunicationsSocket, buf, blkSize, 0)) == SOCKET_ERROR)
	{
		fprintf(stderr, "send: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (!ret)
	{
		fprintf(stderr, "No response from the other side\n");
		return 0;
	}

	if (verbose >= 2)
		printf(".");

	return 1;
}


int main(int argc, char** argv)
{
#ifdef _WIN32
	WSADATA data;
#endif

	int ret;
	unsigned int pos;
	unsigned int fileSize;
	FILE* f;
	char buf[BLKSIZE];
	int arg;
	int nonEscapedArg = 0;
	int help = 0;

	memset(inputName, 0, sizeof(inputName));
	memset(outputHostName, 0, sizeof(outputHostName));
	memset(outputName, 0, sizeof(outputName));

	for (arg = 1; arg < argc; arg++)
	{
		if (!stricmp(argv[arg], "-q") || !stricmp(argv[arg], "--quiet"))
			verbose = 0;
		else if (!stricmp(argv[arg], "-v") || !stricmp(argv[arg], "--verbose"))
			verbose = 2;
		else if (!stricmp(argv[arg], "-h") || !stricmp(argv[arg], "--help"))
			help = 1;
		else
		{
			if (nonEscapedArg == 0)
				strcpy(inputName, argv[arg]);
			else if (nonEscapedArg == 1)
			{
				char* delimiter = strchr(argv[arg], ':');
				if (delimiter)
				{
					memcpy(outputHostName, argv[arg], delimiter - argv[arg]);
					strcpy(outputName, delimiter + 1);
				}
				else
					strcpy(outputHostName, argv[arg]);

				if (!outputName[0])
				{
					int i;
					for (i = strlen(inputName) - 1; i >= 0; i--)
						if (strchr(":/\\", inputName[i]))
							break;

					strcpy(outputName, &inputName[i + 1]);
					
				}
			}
			nonEscapedArg++;
		}
	}

	if (!outputName[0])
		help = 1;

	if (help)
	{
		printf("TCPPut: File send client using plain TCP/IP\n");
		printf("By Mikael Kalms (mikael@kalms.org) 2003-03-03\n");
		printf("Usage: tcpput [options] <source filename> <destination host>:[destination filename]\n");
		printf("Options:\n");
		printf("         -h,--help       Display this help text\n");
		printf("         -q,--quiet      Only output error messages\n");
		printf("         -v,--verbose    Output debug info\n");
		return 0;
	}

#ifdef _WIN32
	if ((ret = WSAStartup( MAKEWORD( 2, 2 ), &data )) == SOCKET_ERROR)
	{
		fprintf(stderr, "WSAStartup: %d\n", WSAGetLastError());
		return 1;
	}

	if (verbose >= 2)
		printf("WSAStartup OK\n");
#endif

	if (!(f = fopen(inputName, "rb")))
	{
		fprintf(stderr, "Unable to open file %s\n", inputName);
		destroySpeaker();
		return 1;
	}

	fseek(f, 0, SEEK_END);
	fileSize = ftell(f);
	fseek(f, 0, SEEK_SET);

	if (!createSpeaker())
		return 0;

	if (verbose >= 2)
		printf("createSpeaker() OK\n");

	if (verbose >= 1)
		printf("Transferring %d bytes from file %s to %s:%s...", fileSize, inputName, outputHostName, outputName);

	{
		unsigned long strLen = htonl(strlen(outputName));
		unsigned long size;
		size = htonl(fileSize);
		if (!sendBlk(&strLen, sizeof(strLen))
			|| !sendBlk(&size, sizeof(size))
			|| !sendBlk(outputName, strlen(outputName)))
		{
			fclose(f);
			destroySpeaker();
			return 0;
		}
	}

	for (pos = 0; pos < fileSize; )
	{
		unsigned int blkSize = BLKSIZE;
		if (pos + blkSize > fileSize)
			blkSize = fileSize - pos;

		fread(buf, blkSize, 1, f);

		if (!sendBlk(buf, blkSize))
		{
			fclose(f);
			destroySpeaker();
			return 0;
		}

		pos += blkSize;
	}

	fclose(f);

	destroySpeaker();

	if (verbose >= 1)
		printf("done\n");
	
	return 0;
}
